%%% -*- indent-tabs-mode: nil -*-
%%%
%%% Author:
%%%   Juergen Zimmer <jzimmer@mathweb.org>
%%%   Andreas Franke <afranke@mathweb.org>
%%%
%%% Copyright:
%%%   Juergen Zimmer, 2001
%%%
%%% Last change:
%%%   $Date: 2001/09/14 08:37:41 $ by $Author: jzimmer $
%%%   $Revision: 1.18 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/**
% This is a generic XML-RPC server that handles XML-remote-procedure-calls.
% It is fully compatible with the Java XML-RPC package of
% Hannes Wallnoefer (see http://www.xmlrpc.com/).
% It handles XMLRPC-requests are of the form:
%
% POST /RPC2 HTTP/1.0
% Content-Type: text/xml
% Content-Length: 100
%
% <methodCall>
%   <methodName>Math.add</methodName>
%   <params>
%     <param>
%       <value><i4>41</i4></value>
%     </param>
%   </params>
% </methodCall>
%
%
% Fault Codes:
%   0 -  9   Reserved for generic exceptions and internal errors.
%            FaultString may contain a printed representation of the exception.
%   0        Can be used for anything. Fallback code.
%   1        Type error.
%    ...
%  10 - 19   Errors related to the message as a whole. Examples:
%            * /PATH is not /RPC2, and no handler can be found for it.
%              (well, alternatively this could be treated as an HTTP 404)
%            * Missing Content-Length header.
%            * Parse error, e.g. message body not well-formed.
%    ...
%  20 - 29   Errors related to the handler.
%  20        Reserved for future use.
%  21        Unknown handler. No such handler has been registered.
%  22        Handler permanently unreachable. A (remote) handler has been
%            registered, but its process is not running any more.
%            [Not used yet.]
%  23        Handler temporarily unreachable. A (remote) handler has been
%            registered, but an attempt to contact its site failed. This error
%            may go away next time.
%            [Not used yet.]
%    ...
%  30 - 39   Errors related to the method name.
%  30        Reserved for future use.
%
%  40 - 49   Errors related to the params.
%  40        Generic params error. (Result of arity mismatch.)
%  41        Wrong number of params.
%  42        Not enough params given. More params were expected.
%  43        Too many params given. Less params were expected.
%  44        Type error. A param has an unexpected type.
%    ...
% 100 - inf  User defined faults.
*/

functor

import
   Error(exceptionToMessage messageToVirtualString registerFormatter)
   Property(put)
   URL(make toVirtualString)
   Browser % this is used only if Debug == true
   System
   
   HttpHandler('class':BaseHTTPHandler)
   at 'x-ozlib://mathweb/httpd/HttpHandler.ozf'
   
   XmlParser
   at 'x-ozlib://mathweb/xmlparser/XmlParser.ozf'
   PModule
   at 'x-ozlib://mathweb/xmlrpc/xmlrpc.parser.ozf'
   SModule
   at 'x-ozlib://mathweb/xmlrpc/xmlrpc.sem.ozf'
   XmlrpcWriter('class':Xmlrpc)
   at 'x-ozlib://mathweb/xmlrpc/xmlrpc.printer.ozf'
   
export
   httpHandler: HTTPHandler
   
define
   VS2A = VirtualString.toAtom
   VS2S = VirtualString.toString
   
   fun {ExceptionToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}} 
   end

   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% fault errors %%%%%%%%%%%%%%%%%%%%%%%%%
   {Error.registerFormatter fault
    fun {$ E}
       T = 'xmlrpc fault' 
    in 
       case E of fault(FaultCode FaultString) then 
          error(kind:T msg:'Internal xmlrpc fault' 
                items:[hint(l:'code' m: FaultCode)
                       hint(l:'message' m: FaultString)
                      ]
               )
       else
          error(kind:T items:[line(oz(E))])
       end 
    end}
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
%   Debug = true
   Debug = false

   proc {DebugBrowse X}
      if Debug then {Browser.browse X} end
   end
   
   {Property.put 'errors.width' 100}
   {Property.put 'errors.depth' 100}
     
   ParserModule = {XmlParser.link PModule SModule}

   {XmlParser.setDebug Debug}
   
   %% type checking routines
   fun {Is1ary Handler}
      {IsObject Handler} orelse
      ({IsProcedure Handler} andthen {Procedure.arity Handler}==1)
   end
   fun {Is3ary Handler}
      {IsProcedure Handler} andthen {Procedure.arity Handler}==3
   end
   proc {CheckHandlerType Handler}
      if {Is1ary Handler} orelse {Is3ary Handler} then skip
      else
         {Exception.raiseError kernel(type 'IsHandler' [Handler]
                                      'procedure/1,3' 1 'CheckHandlerType')}
      end
   end
   
   %% a handler can execute a method call
   %% here are several alternative calling conventions for a handler
   local
      %% Handler is a 2ary function
      proc {ApplyProc Handler MethodName Params ?Result}
         {Handler MethodName Params ?Result}
      end
      %% Handler is an object that provides an execute(...) method
      proc {CallExecute Handler MethodName Params ?Result}
         {Handler execute(MethodName Params ?Result)}
      end
      %% Handler is an object where MethodName should be applied directly
      %% * MethodName(Params ?Result)
      proc {CallSimple Handler MethodName Params ?Result}
         M = {VS2A MethodName} in
         {Handler M(Params ?Result)}
      end
      %% * MethodName(Param1 Param2 ... ?Result)
      proc {CallAsTuple Handler MethodName Params ?Result}
         Msg = {List.toTuple {VS2A MethodName} {Append Params [Result]}} in
         {Handler Msg}
      end
      %% a single param of type <struct> to construct the Oz method call
      %% * MethodName(Name1:Value1 Name2:Value2 ... ?Result)
      local
         fun {GetFirstUnusedIntFeature R Start}
            if {HasFeature R Start}
            then {GetFirstUnusedIntFeature R Start+1}
            else Start
            end
         end
      in
         proc {CallAsRecord Handler MethodName Params ?Result}
            case Params of [Struct=unit(...)] then
               Meth = {AdjoinAt {Adjoin Struct {VS2A MethodName}}
                       {GetFirstUnusedIntFeature Struct 1} Result}
            in
               {Handler Meth}
            else
               %% use CallTuple as a fallback
               {CallAsTuple Handler MethodName Params ?Result}
            end
         end
      end
      %% try CallExecute, CallAsRecord and CallAsTuple
      proc {TryCallObject Handler MethodName Params ?Result}
         try
            {CallExecute Handler MethodName Params ?Result}
         catch error(object(arityMismatch...)...) then
            try 
               {CallAsRecord Handler MethodName Params ?Result}
            catch error(object(arityMismatch...)...) then
               {CallAsTuple Handler MethodName Params ?Result}
            end
         end
      end
   in
      %% map any errors to the according fault codes
      proc {Try Call Handler MethodName Params ?Result}
         try
            {Call Handler MethodName Params ?Result}
         catch error(object(arityMismatch...)...) then
            raise fault(40 "Arity mismatch.") end
         end
      end
      %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% public function for executing a method on a handler
      %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      fun {CallHandler Handler MethodName Params}
         Call = case Handler.mode
                of fun2    then ApplyProc
                [] execute then CallExecute
                [] simple  then CallSimple
                [] tuple   then CallAsTuple
                [] record  then CallAsRecord
                [] object  then TryCallObject
                [] Mode then
                   {Exception.raiseError unknownMode(Mode)} dummy
                end
      in
         {Try Call Handler.1 MethodName Params}
      end
   end
   
   %% guess an appropriate mode
   fun {GuessModeForHandler Handler}
      if {Is3ary Handler} then fun2
      else object end
   end
   
   %% =======================================
   %% a registry for xmlrpc handlers
   %% =======================================

   proc {Skip0} skip end
   %%
   class HandlerRegistry
      
      feat
         handlers
         
      meth init(...) = Init
         self.handlers = {NewDictionary}
      end
      
      %% should multiple handlers be allowed for the same name/path?
      %% (the broker allows this, but I'm not sure if we want this here, too)
      %% if we don't want it, addHandler should fail for a conflicting handler
      %% if you want to replace a handler, you can call removeHandler first
      %% (but then you will be removed to when the first server shuts down,
      %% so this is probably not a solution...)
      meth addHandler(NameOrPath Handler mode:Mode<=_
                      onRemove:OnRemove<=Skip0)
         NameAtom = {VS2A NameOrPath}
         if {Dictionary.member self.handlers NameAtom} then
            {Exception.raiseError xmlrpc(handlerExists NameAtom)}
         end
         if {IsFree Mode} then
            Mode = {GuessModeForHandler Handler}
         end
         {CheckHandlerType Handler}
         Entry = handler(Handler mode:Mode onRemove:OnRemove)
      in
         {Dictionary.put self.handlers NameAtom Entry}
      end
      
      meth removeHandler(NameOrPath)
         NameAtom = {VS2A NameOrPath} in
         %% if the entry does not exist, Dictionary.remove does nothing
         case {CondSelect self.handlers NameAtom notFound}
         of notFound then skip
         [] Entry then
            {Dictionary.remove self.handlers NameAtom}
            OnRemove = {CondSelect Entry onRemove Skip0}
            Done
         in
            thread {OnRemove} Done=unit end
            {WaitOr Done {Alarm 10000}} % signal error ?
         end
      end

      meth removeAllHandlers
         {ForAll {Dictionary.keys self.handlers}
          proc {$ Key}
             {self removeHandler(Key)}
          end}
      end

      meth showHandlers
         {System.showInfo "Registered XML-RPC handlers:"}
         {ForAll {Dictionary.keys self.handlers}
          proc {$ Key}
             {System.showInfo Key}
          end}
      end
   end

   %%
   %% a connection wrapper with support for xml & xmlrpc output 
   %%
   class ConnectionWrapper
      from Xmlrpc

      attr connection

      meth init(Connection)
         connection<-Connection
         %%Xmlrpc, init %% this method doesn't exist
      end
      meth write(vs:VS)=Write
         {@connection Write}
      end
      
      meth writeResponse(Meth)
         {@connection beginResponse(200 contentType:'text/xml')}
         %%{@connection headerLine('content-encoding' 'us-ascii')}
         Xmlrpc, Meth
         {@connection endResponse}
      end
      meth writeParseError()
         {@connection beginResponse(400)}
         {@connection doctype}
         {@connection tag(html(head(title("400 Parse Error"))
                               body(h1("Parse Error")
                                    "Sorry, your request could not be parsed.")))}
         {@connection endResponse}
      end
      
      meth otherwise(M)
         {@connection M}
      end
   end
   
   class Self=HTTPHandler
      from
         BaseHTTPHandler
         HandlerRegistry
         
      feat
         handlers
         
      meth init()
         HandlerRegistry, init
      end

      /*
      meth beginResponseBase
         {Connection beginResponse}
         {Connection tag(base(href:Connection.server.baseURL)#"\n")}
      end
      */

      meth handle(UriInContext Request Connection)
         case {Request getMethodS($)}
         of "GET" then
            {System.showInfo "GET !!!!!"}
            Self, handleGET(UriInContext Request Connection)
         [] "POST" then
            {System.showInfo "POST !!!!!"}
            Self, handlePOST(UriInContext Request Connection)
         [] Method then
            {System.showInfo "unknown method !!!!!"}
            {Connection errMethodNotImplemented(Method {Request getUriS($)})}
         end
      end
      
      /*
      meth 'GET'(X...) = GET
         P = {URL.make X}.path in
         case P of [nil]
         then HttpConnection, 'GET'("/index.html")
         else HttpConnection, GET
         end
      end
      */
      meth handleGET(UriInContext Request Connection)
         if UriInContext == [nil]
         then {System.showInfo "X"} {Connection 'GET'("/index.html")}
         else {System.showInfo "Y"} {Connection 'GET'({Request getUriS($)})}
         end 
      end
      
      meth handlePOST(UriInContext Request Connection) = HANDLE_POST
         Conn    = {New ConnectionWrapper init(Connection)}
         X       = {Request getUriS($)}
         Content = {Request getContentVS($)}
         
         %%{DebugBrowse Content}
         U = {URL.make X}
         U1 = {Record.subtractList U [device scheme authority query fragment]}
         P = {URL.toVirtualString U1}
      in
         {DebugBrowse posted(HANDLE_POST)}
         try
            Result =
            case {VS2S P} of &/|&R|&P|&C|&2|&/|HandlerName then
               %%Self, WriteResponse(Conn writeFault(0 "Please send HTTP with path /RPC2"))
               %%
               %% for paths other than "/RPC2",
               %% the path itself should be a registered handler name
               %%
               case {Dictionary.condGet self.handlers {VS2A HandlerName} unit}
               of unit then 
                  {Conn errNotFound}
                  raise close end %% dummy
               [] Handler then
                  MethodCall = Self, ParseRequest(Content $ connection:Conn) in
                  {CallHandler Handler MethodCall.methodName MethodCall.params}
               end
            [] "/RPC2" then
               %%
               %% if the path is "/RPC2"
               %% then <methodName>'s content should be HandlerName.MethodName
               %%
               MethodCall = Self, ParseRequest(Content $ connection:Conn)
               %%{DebugBrowse MethodCall}
               HandlerName MethodName
            in
               {String.token MethodCall.methodName &. ?HandlerName ?MethodName}
               case {Dictionary.condGet self.handlers {VS2A HandlerName} unit}
               of unit then
                  raise fault(1 "No handler available for "#HandlerName#".")
                  end
               [] Handler then
                  %%RightCall = {List.toTuple {VS2A MethodName}
                  %%             {Append Params [Result]}}
                  %%{DebugBrowse RightCall}
                  {CallHandler Handler MethodName MethodCall.params}
               end
            else
               {Conn errNotFound}
               raise close end %% dummy
            end
         in
            {Conn writeResponse(writeResult(Result))}
         catch Err then
            {System.showInfo {ExceptionToVS Err}}
            case Err of close then
               raise close end
            [] fault(FaultCode FaultString) then
               {Conn writeResponse(writeFault(FaultCode FaultString))}
            else
               {Conn writeResponse(writeFault(0 "Some Error occured..."#
                                              {ExceptionToVS Err}))}
            end
         end
         %%
         {Delay 1000}
         raise close end
      end
      
      meth ParseRequest(Request ?MethodCall connection:Conn)
         Accept in
         {XmlParser.parseVS Request ParserModule
          'start_methodCall'(?MethodCall) ?Accept}
         if Accept==false then
            {Conn writeParseError}
            {Delay 1000}
            raise close end
         end
      end
   end

end



